@@ -6,6 +6,8 @@ from google.protobuf import any_pb2, message, timestamp_pb2
 from .proto.cosmos.authz.v1beta1 import authz_pb2 as cosmos_authz_pb
 from .proto.cosmos.authz.v1beta1 import tx_pb2 as cosmos_authz_tx_pb
 
+from .proto.injective.exchange.v1beta1 import authz_pb2 as injective_authz_pb
+
 from .proto.cosmos.bank.v1beta1 import tx_pb2 as cosmos_bank_tx_pb
 from .proto.cosmos.base.v1beta1 import coin_pb2 as cosmos_base_coin_pb
 
@@ -19,14 +21,14 @@ from .proto.injective.peggy.v1 import msgs_pb2 as injective_peggy_tx_pb
 
 from .proto.injective.oracle.v1beta1 import tx_pb2 as injective_oracle_tx_pb
 
+from .proto.cosmos.staking.v1beta1 import tx_pb2 as cosmos_staking_tx_pb
+
+from .proto.cosmos.distribution.v1beta1 import tx_pb2 as cosmos_distribution_tx_pb
+
 from .constant import Denom
 from .utils import *
 from typing import List
 
-from .wallet import Address
-
-from web3 import Web3, HTTPProvider
-
 class Composer:
     def __init__(self, network: str):
         self.network = network
@@ -383,7 +385,7 @@ class Composer:
             bid_amount=self.Coin(amount=be_amount, denom="inj"),
         )
 
-    def MsgGrant(self, granter: str, grantee: str, msg_type: str, expire_in: int):
+    def MsgGrantGeneric(self, granter: str, grantee: str, msg_type: str, expire_in: int):
         auth = cosmos_authz_pb.GenericAuthorization(msg=msg_type)
         any_auth = any_pb2.Any()
         any_auth.Pack(auth, type_url_prefix="")
@@ -397,6 +399,65 @@ class Composer:
             granter=granter, grantee=grantee, grant=grant
         )
 
+    def MsgGrantTyped(self, granter: str, grantee: str, msg_type: str, expire_in: int, subaccount_id: str, **kwargs):
+        auth = None
+        if msg_type == "CreateSpotLimitOrderAuthz":
+            auth = injective_authz_pb.CreateSpotLimitOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "CreateSpotMarketOrderAuthz":
+            auth = injective_authz_pb.CreateSpotMarketOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "BatchCreateSpotLimitOrdersAuthz":
+            auth = injective_authz_pb.BatchCreateSpotLimitOrdersAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "CancelSpotOrderAuthz":
+            auth = injective_authz_pb.CancelSpotOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "BatchCancelSpotOrdersAuthz":
+            auth = injective_authz_pb.BatchCancelSpotOrdersAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "CreateDerivativeLimitOrderAuthz":
+            auth = injective_authz_pb.CreateDerivativeLimitOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "CreateDerivativeMarketOrderAuthz":
+            auth = injective_authz_pb.CreateDerivativeMarketOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "BatchCreateDerivativeLimitOrdersAuthz":
+            auth = injective_authz_pb.BatchCreateDerivativeLimitOrdersAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "CancelDerivativeOrderAuthz":
+            auth = injective_authz_pb.CancelDerivativeOrderAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "BatchCancelDerivativeOrdersAuthz":
+            auth = injective_authz_pb.BatchCancelDerivativeOrdersAuthz(
+                subaccount_id=subaccount_id, market_ids=kwargs.get("market_ids")
+            )
+        elif msg_type == "BatchUpdateOrdersAuthz":
+            auth = injective_authz_pb.BatchUpdateOrdersAuthz(
+                subaccount_id=subaccount_id, spot_markets=kwargs.get("spot_markets"), derivative_markets=kwargs.get("derivative_markets")
+            )
+
+        any_auth = any_pb2.Any()
+        any_auth.Pack(auth, type_url_prefix="")
+
+        grant = cosmos_authz_pb.Grant(
+            authorization=any_auth,
+            expiration=timestamp_pb2.Timestamp(seconds=(int(time()) + expire_in)),
+        )
+
+        return cosmos_authz_tx_pb.MsgGrant(
+            granter=granter, grantee=grantee, grant=grant
+        )
+
     def MsgExec(self, grantee: str, msgs: List):
         any_msgs: List[any_pb2.Any] = []
         for msg in msgs:
@@ -447,59 +508,31 @@ class Composer:
             bridge_fee=self.Coin(amount=be_bridge_fee, denom=peggy_denom)
         )
 
-    def SendToCosmos(self, ethereum_endpoint: str, private_key: str, token_contract: str, receiver: str, amount: int, maxFeePerGas: int,  maxPriorityFeePerGas: int, peggo_abi: str, decimals=18):
-        if self.network == 'testnet':
-            peggy_proxy_address = "0xd6Da9dA014806Fdb64bF39b48fcA386AE3420d21"
-        elif self.network == 'mainnet':
-            peggy_proxy_address = "0xF955C57f9EA9Dc8781965FEaE0b6A2acE2BAD6f3"
-        else:
-            print("SendToCosmos is only supported on Mainnet & Testnet")
-        web3 = Web3(Web3.HTTPProvider(ethereum_endpoint))
-        contract = web3.eth.contract(address=peggy_proxy_address, abi=peggo_abi)
+    def MsgDelegate(
+        self,
+        delegator_address: str,
+        validator_address: str,
+        amount: float
+    ):
 
-        token_contract_address = web3.toChecksumAddress(token_contract)
+        be_amount = amount_to_backend(amount, 18)
 
-        receiver_address = Address.from_acc_bech32(receiver)
-        receiver_ethereum_address = Address.get_ethereum_address(receiver_address)
-        receiver_address_checksum = web3.toChecksumAddress(receiver_ethereum_address)
-        receiver_slice = receiver_address_checksum[2:]
-        receiver_padded_address = '0x' + receiver_slice.zfill(64)
+        return cosmos_staking_tx_pb.MsgDelegate(
+            delegator_address=delegator_address,
+            validator_address=validator_address,
+            amount=self.Coin(amount=be_amount, denom="inj"),
+        )
 
-        destination = web3.toBytes(hexstr=receiver_padded_address)
+    def MsgWithdrawDelegatorReward(
+        self,
+        delegator_address: str,
+        validator_address: str
+    ):
 
-        sender_ethereum_address = web3.eth.account.privateKeyToAccount(private_key).address
-        sender_address_checksum = web3.toChecksumAddress(sender_ethereum_address)
-        nonce = web3.eth.get_transaction_count(sender_address_checksum)
-
-        amount_to_send = int(amount * pow(10, decimals))
-
-        gas = contract.functions.sendToCosmos(
-            token_contract_address,
-            destination,
-            amount_to_send,
-        ).estimateGas({'from': sender_address_checksum})
-
-        transaction_body = {
-            'nonce': nonce,
-            'gas': gas,
-            'maxFeePerGas': web3.toWei(maxFeePerGas, 'gwei'),
-            'maxPriorityFeePerGas': web3.toWei(maxPriorityFeePerGas, 'gwei'),
-        }
-
-        tx = contract.functions.sendToCosmos(
-            token_contract_address,
-            destination,
-            amount_to_send,
-        ).buildTransaction(transaction_body)
-
-        signed_tx = web3.eth.account.signTransaction(tx, private_key=private_key)
-
-        try:
-            tx_hash = web3.eth.send_raw_transaction(signed_tx.rawTransaction)
-            print("Transferred {} {} from {} to {}".format(amount, token_contract, sender_ethereum_address, receiver))
-            print("Transaction hash:", web3.toHex(tx_hash))
-        except Exception as e:
-            print("Transaction failed", e)
+        return cosmos_distribution_tx_pb.MsgWithdrawDelegatorReward(
+            delegator_address=delegator_address,
+            validator_address=validator_address
+        )
 
     # data field format: [request-msg-header][raw-byte-msg-response]
     # you need to figure out this magic prefix number to trim request-msg-header off the data
